module Calculations(Calculation,Step,Hint(..),conclusion,result,calcsteps,
                    prettyCalc,shortprettyCalc,htmlCalc
                   ) where

import List
import Utils
import UU_Pretty (indent)
import UU_Pretty_ext
import PPrint
import Expressions
import SubExpressions
import Substitutions
import State
import Names
import Laws
import Html

--type Law         = CondEq
type Law         = String
type Calculation = (Exp,[Step],[Hint])
type Step        = (Hint, Exp)
data Hint        = Direct Law Subst | Indirect Law Subst [Calculation]

conclusion :: Calculation -> Exp
conclusion (exp,steps,_) = if null steps then exp
                           else result (last steps)

calcsteps :: Calculation -> [Step]
calcsteps (_,steps,_) = steps

result :: Step -> Exp
result (_,exp) = exp

lawname :: Law -> String
--lawname (n,_,_) = n
lawname n = n

-- default way of printing out a Calculation - with side calculations 
-- for conditional rules

prettyCalc :: Int -> Int -> Calculation -> String
prettyCalc w i (x,ss,_) = prettyLine w i x ++ "\n" 
                             ++ sepby "\n" (map (prettyStep w i) ss) 
prettyLine w i x        = wdisplay w $ indent (i+3) $ ppExp 0 $ prepare x

prepare x = nearbeta (fst (evalContext (x,[]) (fullcse (x,[]))))

prettyStep w i (h,x)  = take i spaces ++ "= { "  ++ 
                         prettyHint w i h++ " }\n" ++ prettyLine w i x 
prettyHint w i (Direct why _)      = lawname why 
prettyHint w i (Indirect why _ cs) = 
         sepby "\n\n"(lawname why: map (prettyCalc w (i+3))cs) 
                   ++ "\n" ++ take i spaces ++ " "
       where spaces = ' ' : spaces



-- leave out the side calculations

shortprettyCalc :: Int -> Int -> Calculation -> String
shortprettyCalc w i (x,ss,_) = prettyLine w i x ++ "\n" 
                              ++ sepby "\n" (map (shortprettyStep w i) ss)
                              ++ if null ss then "" 
                                 else prettyLine w i (result (last ss))
shortprettyStep w i (h,_)  = take i spaces ++ "= { "  ++ 
                               prettyHint w i h++ " }\n"


-- debug output in HTML format (with expander buttons)

htmlCalc :: Calculation -> Html

type IDSupply = [String]

type ID a = St IDSupply a

freshID :: St IDSupply String
freshID = St (\(id:ids) -> (id,ids))

idsupply = tail (concat (iterate next [""]))
       where next xs = [letter:x | letter <- ['a'..'z'], x <- xs]


htmlCalc calc = header << 
                  (thetitle << "MAG Debugging Output"
                    +++ itag "LINK" ! [rel "STYLESHEET", 
                                       thetype "text/css",
                                       href "mag.css"])
                  +++ body << (expanderscript 
                                  +++  eval idsupply (htmlCalc' calc))

htmlCalc' :: Calculation -> ID Html

htmlCalc' (x,ss,fails) = do id <- freshID
                            steps <- mapM htmlStep ss
                            failures <- mapM htmlHint fails
                            failures' <- if null fails then return noHtml
                                         else expander False "calc"
                                               (hr +++ map (+++hr) failures)
                                               (italics << "failures")
                            exp <- expander True "calc" 
                                    (htmlLine x 
                                       +++ foldl (+++) noHtml steps
                                       +++ failures'
                                       +++ br)
                                    br
                            return exp

htmlLine x = thediv ! [theclass "expression"] 
               << p << (linesToHtml $ lines $ display $ ppExp 0 $ prepare $ x)

htmlStep :: Step -> ID Html

htmlStep (h,x) = 
   do hint <- htmlHint h 
      return $ hint +++ htmlLine x

htmlHint :: Hint -> ID Html

htmlHint (Direct why s) =
    do substHtml <- htmlSubst' s
       lawHtml <- htmlLaw why
       return $ thediv ! [theclass "hint"] 
                 << p << ("= { " +++ lawHtml +++ " }" +++ substHtml)

htmlHint (Indirect why s cs) = 
  do substHtml <- htmlSubst' s
     subcalcs <- mapM htmlCalc' cs
     detailsHtml <- expander True "calc" 
                     (thediv ! [theclass "subcalc"]
                        << foldl (+++) noHtml subcalcs)
                     (italics << "details")
     lawHtml <- htmlLaw why
     return $
       thediv ! [theclass "hint"]
         <<  ("= { " +++ lawHtml +++ " }" +++ substHtml +++ detailsHtml)

htmlLaw :: Law -> ID Html
htmlLaw name = return (name +++ noHtml)
{-
htmlLaw (name,conds,law) = 
   do exp <- expander False "calc" details (italics << "law")
      return (name +++ exp)
    where details = htmlEquation law +++ if null conds then noHtml 
                                         else conddetails
          conddetails = p << bold << "if" +++ map htmlEquation conds
-}
htmlEquation :: Equation -> Html

htmlEquation eqn = thediv ! [theclass "expression"] 
                     << p << (linesToHtml $ lines $ display $ ppEqn $ eqn)

htmlSubst' s = expander False "calc" 
                     (thediv ! [theclass "subcalc"] << htmlSubst s) 
                     (italics << "substitution")

expanderscript = tag "SCRIPT" ! [thetype "text/javascript"]
                   << Html [MarkupString (unlines [
 "<!--",
 "function dobranch(branchon, branchoff, imageID) {",
 "         if (branchon.style.display == \"none\") {",
 "                branchon.style.display = \"\";",
 "                branchoff.style.display = \"none\";",
 "                imageID.src = \"minus.gif\";",
 "       } else {",
 "                branchon.style.display = \"none\";",
 "                branchoff.style.display = \"\";",
 "                imageID.src = \"plus.gif\";",
 "       }",
 "}",
 "// -->"])]

expander :: (HTML a,HTML b) => Bool -> String -> a -> b -> ID Html

expander ison classname onhtml offhtml =
  do id <- freshID
     return $
       tag "SPAN" ! [theclass "expander",
                                       strAttr "ONCLICK" 
                                        ("dobranch(divon"++id++
                                                   ",divoff"++id++
                                                   ",img"++id++")")
                                  ]
                         << image ! [src (if ison then "minus.gif" 
                                                  else "plus.gif"),
                                     identity ("img"++id),
                                     alt "button"]
                        +++ tag "SPAN" ! [theclass classname,
                                          identity ("divon"++id),
                                          style ison]
                          << onhtml
                        +++ tag "SPAN" ! [theclass classname,
                                          identity ("divoff"++id),
                                          style (not ison)]
                          << offhtml
            where style True = thestyle "display:"
                  style False = thestyle "display: none"

