> module ReferenceMatching
>                 (simplematch,  simplematch',
>                  onestepmatch, onestepmatch',
>                  twostepmatch, twostepmatch',
>	           convert,convert',anticonvert,
>                  convertSubst,anticonvertSubst)
>
> where

> import Monad
> import List
> import Maybe

> import Utils (collect,subsplus,subseteq)
> import State
> import Names hiding (freshVar)
> import qualified Names
> import Types hiding (apply)
> import qualified Types
> import Constants
> import qualified Expressions as O
> import qualified Substitutions as OS
> import FiniteMaps

> import PPrint

import Trace

> showExp = prettyExp.anticonvert

> simplematch e1 e2 
>    = evalSt ([],[]) $
>          matches simpleappsplit [Rule (e1,e2)]

> simplematch' s (p,bs) (t,cs)
>    = evalSt ([],[]) $
>       do match <- matches simpleappsplit [Rule (p',t')]
>          return $ (`compose` s) $ funcom (thaw cs') $ match
>       where bs' = map fst bs
>             cs' = map fst cs
>             p' = apply s $ freeze bs' p
>             t' = apply s $ freeze cs' t

> match appsplitter p (t,bs) = match' appsplitter idSubst (p,[]) (t,bs)

> match' appsplitter s (p,bs) (t,cs) =
>     forkName $
>      do match <- matches appsplitter [Rule (p',t')]
>         return $ (`compose` s) $ funcom (thaw cs') $ match
>       where bs' = map fst bs
>             cs' = map fst cs
>             p' = apply s $ freeze bs' p
>             t' = apply s $ freeze cs' t

> onestepmatch = match (tuplifysplitter onestepappsplit)
> onestepmatch' = match' (tuplifysplitter onestepappsplit)

> twostepmatch p t = if twostepvalid p 
>                    then match (tuplifysplitter twostepappsplit) p t
>                    else match (tuplifysplitter onestepappsplit) p t

> twostepmatch' s (p,bs) t = 
>    if twostepvalid p 
>    then match' (tuplifysplitter twostepappsplit) s (p,bs) t
>    else match' (tuplifysplitter onestepappsplit) s (p,bs) t

Checking whether a pattern is flexible

> flex (Ap f _ _) = flex f
> flex (Tupl _ _) = False
> flex (PVar _ _) = True
> flex (LVar _ _) = False
> flex (Con  _ _) = False
> flex (Lam _ _ _) = True


Checking whether a pattern is valid for two step matching

> twostepvalid :: Exp -> Bool
> twostepvalid (Lam _ b _ ) = twostepvalid b
> twostepvalid (Ap f a _) = (not (flex f) || validarg a)
>			                  && twostepvalid f
>			                  && twostepvalid a
>         where validarg exp = 
>                    null (freevars body)
>		  && null (args \\ unboundlocals body)
>                 && (not (null (constants body) 
>                             && null (unboundlocals body \\ args)))
>                       where (args,body) = getab exp []
>                             getab (Lam p b _) xs = getab b (lamvars p++xs)
>                             getab e xs = (xs,e)
> twostepvalid (Tupl es _) = and $ map twostepvalid es
> twostepvalid _ = True

> type NameL a = NameT [] a

> freshVar :: NameL VarId
> freshVar = Names.freshVarT

The datatype of expressions we'll use through this module

> data Exp = LVar VarId Type | PVar VarId Type
>          | Con Constant Type
>          | Ap Exp Exp Type | Lam LamBind Exp Type
>          | Tupl [Exp] Type
>     deriving Show

> data LamBind = LamVar VarId | LamTup [LamBind]
>     deriving Show

.. and substitutions

> type Subst = (FiniteMap VarId Exp,Sub)

> idSubst :: Subst
> idSubst = (zeroFM,idsub)

Make this a nice operator? Unfortunately can't use :=

> mapvar :: VarId -> Exp -> Subst
> p `mapvar` e = (unitFM p e,idsub)

> singletype :: Type -> Type -> [Subst]
> singletype s t = [(zeroFM, st) | st <- tmatch s t]

> compose :: Subst -> Subst -> Subst
> s1@(se1,st1) `compose` (se2,st2) = (mapFM (const (apply s1)) se2 `joinFM` se1
>			             ,comp st1 st2)

> funcom :: (Exp -> Exp) -> Subst -> Subst
> funcom f (se,st) = (mapFM (const f) se,st)

We'll need to convert from the expressions used elsewhere:

> convert :: O.Exp -> Exp
> convert oexp = convert' [] oexp

> convert' bounds (O.Var x t) = if x `elem` bounds 
>                                       then LVar x t
>                                       else PVar x t
> convert' _      (O.Con c t) = Con c t
> convert' bounds oexp@(O.Ap oexp1 oexp2 t) 
>                = maybe apexp maketuple (gettuple oexp)
>     where apexp = (Ap (convert' bounds oexp1) (convert' bounds oexp2) t)
>           maketuple oexps = Tupl (map (convert' bounds) oexps) t

> convert' bounds (O.Lam opat oexp t) 
>         = Lam pat (convert' (newbounds++bounds) oexp) t
>     where (pat,newbounds) = process (convert opat)
>           process (PVar x _) = (LamVar x,[x])
>           process (Tupl es _) = (LamTup tuplpats, concat tuplbounds)
>               where (tuplpats,tuplbounds) = unzip (map process es)
>           process _ = error "can't handle non-variable/tuple lambda bindings"

> gettuple :: O.Exp -> Maybe [O.Exp]

> gettuple exp = gettuple' [] exp
> gettuple' xs (O.Ap exp1 exp2 _) = gettuple' (exp2:xs) exp1
> gettuple' xs (O.Con (Tuple n) _) = if n==length xs then Just xs else Nothing
> gettuple' _ _ = Nothing

> lamvars :: LamBind -> [VarId]
> lamvars (LamVar x) = [x]
> lamvars (LamTup ps) = concat (map lamvars ps)

> lamlammatch :: LamBind -> LamBind -> Maybe [(VarId,VarId)]
> lamlammatch (LamVar x) (LamVar y) = Just [(x,y)]
> lamlammatch (LamTup xs) (LamTup ys) 
>     = if length xs /= length ys then Nothing 
>       else foldr accum (Just []) $ map (uncurry lamlammatch) $ zip xs ys
>   where accum (Just newmatches) (Just matches) = Just (newmatches++matches)
>         accum _ _ = Nothing
> lamlammatch _ _ = Nothing

> lamexpmatch :: LamBind -> Exp -> Maybe [(VarId,Exp)]
> lamexpmatch (LamVar x) e = Just [(x,e)]
> lamexpmatch (LamTup xs) (Tupl es _) 
>    = if length xs /= length es then Nothing
>      else foldr accum (Just []) $ zipWith lamexpmatch xs es
>   where accum (Just newmatches) (Just matches) = Just (newmatches++matches)
>         accum _ _ = Nothing
> lamexpmatch _ _ = Nothing

> lamconvert :: (VarId -> Type -> a) -> ([a] -> Type -> a) 
>                    -> Type -> LamBind -> a

> lamconvert var tup typ (LamVar x) = var x typ
> lamconvert var tup typ (LamTup ps)
>   = tup (map (uncurry $ lamconvert var tup) $ zip (gettupletypes typ) ps) typ

..and back

> anticonvert :: Exp -> O.Exp
> anticonvert (LVar x t) = O.Var x t
> anticonvert (PVar p t) = O.Var p t
> anticonvert (Con c t) = O.Con c t
> anticonvert (Ap e1 e2 t) = O.Ap (anticonvert e1) (anticonvert e2) t

> anticonvert (Lam p e t) 
>    = O.Lam (lamconvert O.Var (\es _ -> makeotuple es) (argtype t) p) 
>            (anticonvert e) t

> anticonvert (Tupl es t) = makeotuple (map anticonvert es)

> makeotuple oexps = foldl1 O.apl (tuplecon:oexps)
>       where tuplecon = O.Con (Tuple (length oexps)) tt
>             tt = arrows (ts ++ [TCon ("Tuple"++show (length oexps)) ts])
>             ts = map O.exptype oexps

> gettupletypes (TCon ('T':'u':'p':'l':'e':_) ts) = ts

And the same for substitutions:

> convertSubst :: OS.Subst -> Subst
> convertSubst (oesub,tsub) =  (mapFM (const convert) $ listToFM oesub,tsub)

> anticonvertSubst :: Subst -> OS.Subst
> anticonvertSubst (esub,tsub) 
>        = (toListFM $ mapFM (const anticonvert) esub,tsub)

What are the unbound local variables used in this expression?

> unboundlocals :: Exp -> [VarId]
> unboundlocals (LVar x _) = [x]
> unboundlocals (PVar _ _) = []
> unboundlocals (Con _ _) = []
> unboundlocals (Ap e1 e2 _) = unboundlocals e1 ++ unboundlocals e2
> unboundlocals (Tupl es _) = concat (map unboundlocals es)
> unboundlocals (Lam p e _) 
>              = filter (\y -> not (y `elem` (lamvars p))) (unboundlocals e) 

What are the free variables in this expression?

> freevars :: Exp -> [VarId]
> freevars (LVar _ _) = []
> freevars (PVar x _) = [x]
> freevars (Con _ _) = []
> freevars (Ap e1 e2 _) = freevars e1 ++ freevars e2
> freevars (Tupl es _) = concat (map freevars es)
> freevars (Lam _ e _) = freevars e

And the constants?

> constants :: Exp -> [Constant]
> constants (LVar _ _) = []
> constants (PVar _ _) = []
> constants (Con c _) = [c]
> constants (Ap e1 e2 _) = constants e1 ++ constants e2
> constants (Tupl es _) = concat (map constants es)
> constants (Lam _ e _) = constants e

Rename bound variable x to y

> renamebound :: VarId -> VarId -> Exp -> Exp
> renamebound x y (LVar z t) = if z==x then LVar y t else LVar z t
> renamebound _ _ (PVar p t) = PVar p t
> renamebound _ _ (Con c t) = Con c t
> renamebound x y (Ap e1 e2 t) = Ap (renamebound x y e1) (renamebound x y e2) t
> renamebound x y (Tupl es t) = Tupl (map (renamebound x y) es) t
> renamebound x y (Lam p e t) = Lam p (renamebound x y e) t

Some operations on Exp

> exptype (LVar _ t) = t
> exptype (PVar _ t) = t
> exptype (Con _ t) = t
> exptype (Ap _ _ t) = t
> exptype (Tupl _ t) = t
> exptype (Lam _ _ t) = t

> apl f a = Ap f a (restype (exptype f))
> lambda (x,t) b = Lam (LamVar x) b (arrow t (exptype b))
> tuplelambda [] b = b
> tuplelambda [(p,t)] b = Lam p b (arrow t (exptype b))
> tuplelambda ps b = Lam (LamTup (map fst ps)) b (arrow tupletype (exptype b))
>          where tupletype = TCon ("Tuple"++show (length ps)) (map snd ps)

> instance Eq Exp where
>   e1 == e2 = eq [] [] (etaContract e1) (etaContract e2)
>      where eq xs ys (LVar x s) (LVar y t) 
>                  = s==t && findvar xs x == findvar ys y
>            eq xs ys (PVar p s) (PVar q t) = p==q && s==t
>            eq xs ys (Con c s)  (Con d t)  = c==d && s==t
>            eq xs ys (Ap e1 e2 s) (Ap f1 f2 t) = s==t &&
>                                                  eq xs ys e1 f1 &&
>                                                  eq xs ys e2 f2
>            eq xs ys (Tupl es s) (Tupl fs t) 
>                    = s==t && length es == length fs &&
>                      (and $ map (uncurry (==)) $ zip es fs)
>            eq xs ys (Lam p e s) (Lam q f t) 
>                    = s==t && case lamlammatch p q of
>                              Just matches -> let (xs',ys') = unzip matches
>                                              in eq (xs'++xs) (ys'++ys) e f
>                              Nothing -> False
>            eq _ _ _ _ = False
>            findvar [] y = Right y
>            findvar (x:xs) y = if x==y then Left 0
>                             else case findvar xs y of 
>                               Left n -> Left (n+1)
>                               Right w -> Right w

> etaredex (Lam p (Ap f exp _) _)
>    = case lamexpmatch p exp of
>      Just matches -> (and $ (map check) $ matches)
>                      && null (map fst matches `intersect` unboundlocals f)
>      Nothing -> False
>  where check (x,LVar y _) = x==y
>        check (_,_) = False
> etaredex e = False

etaRed strips off a top-level lambda abstraction

> etaRed e@(Lam p (Ap f exp _) _) = if etaredex e then f else e
> etaRed e = e

etaContract strips off the lot

> etaContract e@(Lam p b t) = etaRed (Lam p (etaContract b) t)
> etaContract (Ap e1 e2 t) = Ap (etaContract e1) (etaContract e2) t
> etaContract (Tupl es t) = Tupl (map etaContract es) t
> etaContract e = e

> etaNormal :: Exp -> Bool
> etaNormal e@(Lam _ b _) = not (etaredex e) && etaNormal b
> etaNormal (Ap e1 e2 _) = etaNormal e1 && etaNormal e2
> etaNormal (Tupl es _) = and (map etaNormal es)
> etaNormal e = True

> betaNormal :: Exp -> Bool

"(\(x,y) -> x+y) z" *is* beta-normal. 

> betaNormal (Ap (Lam pat _ _) exp _) = isJust (lamexpmatch pat exp)
> betaNormal (Ap e1 e2 _) = betaNormal e1 && betaNormal e2
> betaNormal (Lam x b _) = betaNormal b
> betaNormal e = True

> normal e = etaNormal e && betaNormal e

> freeze :: [VarId] -> Exp -> Exp
> freeze ids (LVar x t) = if x `elem` ids then Con (Constr x) t
>                                         else LVar x t
> freeze ids (PVar p t) = PVar p t
> freeze ids (Con c t) = Con c t
> freeze ids (Ap e1 e2 t) = Ap (freeze ids e1) (freeze ids e2) t
> freeze ids (Tupl es t) = Tupl (map (freeze ids) es) t
> freeze ids (Lam p b t) = Lam p (freeze ids b) t

> thaw :: [VarId] -> Exp -> Exp
> thaw ids (Con (Constr x) t) = if x `elem` ids then LVar x t
>                                               else Con (Constr x) t
> thaw ids (PVar p t) = PVar p t
> thaw ids (LVar x t) = LVar x t
> thaw ids (Con c t) = Con c t
> thaw ids (Ap e1 e2 t) = Ap (thaw ids e1) (thaw ids e2) t
> thaw ids (Tupl es t) = Tupl (map (thaw ids) es) t
> thaw ids (Lam p b t) = Lam p (thaw ids b) t

We'll want to apply substitutions to various types, so:

> class Substitutable a where
>       apply :: Subst -> a -> a

> instance Substitutable a => Substitutable [a] where
>       apply subst = map (apply subst)


> instance Substitutable Exp where
>       apply (se,st) (PVar p t) = lookupDftFM se (PVar p (Types.apply st t)) p
>       apply (se,st) (LVar x t) = LVar x (Types.apply st t)
>       apply (se,st) (Con c t) = Con c (Types.apply st t)
>       apply s@(se,st) (Ap e1 e2 t) = Ap (apply s e1) (apply s e2) 
>                                                 (Types.apply st t)
>       apply s@(se,st) (Tupl es t) = Tupl (map (apply s) es) 
>                                                 (Types.apply st t)                                                
>       apply s@(se,st) (Lam p e t) 
>             = Lam p (apply s e) (Types.apply st t)

OK, here's the interesting code...

A Rule is simply a pair of expressions, the first representing the pattern
and the second the term.

> newtype Rule = Rule (Exp,Exp)

In a Rule, the term must be closed.

> instance Substitutable Rule where
>       apply phi (Rule (p,t)) = Rule (apply phi p,t)

An "appsplitter" takes a boolean indicating whether the function part
of a pattern is flexible, the argument part of a pattern, and a term.
It non-deterministically returns a possible way of splitting up
the term into an application; an expression for the function part of the term
and a list of rules that will resolve matching the given argument pattern
and rules for the argument part.

> type AppSplitter = Bool -> Exp -> Exp -> NameL (Exp, [Rule])

Various parts of the evaluation of matches will require fresh names,
so the entire computation is done under the "NameL" monad. This combines
a name supply with non-determinism (using monad transformers)

> matches :: AppSplitter -> [Rule] -> NameL Subst

> matches _ [] = return idSubst

> matches appsplitter (x:xs) 
>               = do (sigma, ys) <- expresolve appsplitter x
>                    phi <- matches appsplitter (apply sigma (xs++ys))
>                    return (phi `compose` sigma)

If we used a monad instead of a monad transformer, 
matches would look like this, which is rather less pleasant.

 matches (x:xs) = do resolves <- resolve x
                     liftM concat $ sequence $ 
                      do (sigma, ys) <- resolves
                         return $ 
                          do submatches <- matches (apply sigma (xs++ys))
                             return $ 
                              do phi <- submatches
                                 return (phi `compose` sigma)

Since our expressions have types, a rule naturally breaks down
into the two problems of matching the substance of the expressions
and of matching the types. Matching the types isn't important to
our presentation, so lets shuffle it off to one side:

> expresolve :: AppSplitter -> Rule -> NameL (Subst, [Rule])

> expresolve appsplitter rule@(Rule (e1,e2)) = 
>      if not (null (freevars e2)) 
>       then error ("free variables in term: " ++ showExp e2)
>       else do tau <- liftSt $ singletype (exptype e1) (exptype e2)
>               (phi,rules) <- resolve appsplitter rule
>               return (phi `compose` tau, apply tau rules)

> resolve :: AppSplitter -> Rule -> NameL (Subst, [Rule])

> resolve _ (Rule (LVar x _,LVar y _)) 
>    = liftSt $
>        if x==y then [(idSubst,[])]
>                else []

> resolve _ (Rule (PVar p _,e))
>    = liftSt $
>        if null (unboundlocals e) 
>        then [(p `mapvar` e,[])]
>        else []

> resolve _ (Rule (Con c _,Con d _))
>    = liftSt $
>        if c==d then [(idSubst,[])]
>                else []

> resolve _ (Rule (Lam p e _,Lam q f _))
>    = case lamlammatch p q of
>       Just matches -> return (idSubst,[Rule (e,ren matches f)])
>       Nothing -> liftSt []
>  where ren matches f = foldr (\(x,y) exp -> renamebound y x exp) f matches

> resolve _ (Rule (Lam p e s,f))
>    = return (idSubst,[Rule (e,apl f pcopy)])
>  where pcopy = lamconvert LVar Tupl (argtype s) p

> resolve _ (Rule (Tupl es _, Tupl fs _))
>    = if length es == length fs 
>      then return (idSubst,map Rule $ zip es fs)
>      else liftSt []

> resolve appsplitter (Rule (Ap f a _,e))
>    = do (func,argrules) <- appsplitter (flex f) a e
>         return (idSubst,Rule (f,func):argrules)

> resolve _ _ = liftSt []

> simpleappsplit :: AppSplitter

> simpleappsplit _ a (Ap f' a' _) = return (f',[Rule (a,a')])
> simpleappsplit _ _ _ = liftSt []

> onestepappsplit :: AppSplitter

> onestepappsplit fflex a t = 
>      simpleappsplit fflex a t
>       `mplus`
>      do (f',a') <- apps t
>         return (f',[Rule (a,a')])
>       `mplus`
>      do x <- freshVar
>         return (lambda (x,exptype a) t,[])


> apps :: Exp -> NameL (Exp, Exp)
> apps term 
>           = do x <- freshVar
>                (subexp, locs) <- liftSt $ collect $ filter okse 
>                                                      $ subexps term
>                locs' <- liftSt $ subsplus locs
>                let t = exptype subexp
>                let body = replaces term locs' (LVar x t)
>                let fun = lambda (x,t) body
>                if normal fun then return (fun, subexp)
>                              else liftSt []
>     where okse (subexp,_) = unboundlocals subexp `subseteq` oklocals
>           oklocals = unboundlocals term

> data Direction = First | Second | Body | TupElem Int
>            deriving (Eq,Show)
> type Location = [Direction]
> type SubExp = (Exp, Location)

> subexps (LVar x t) = [(LVar x t,[])]
> subexps (PVar p t) = [(PVar p t,[])]
> subexps (Con c t) = [(Con c t,[])]
> subexps (Lam p e t) = [(Lam p e t,[])] ++ map body (subexps e)
>                         where body (s,l) = (s,Body:l)
> subexps (Ap f a t) = [(Ap f a t,[])] ++ map first (subexps f) 
>                                      ++ map second (subexps a)
>                         where first (s,l) = (s,First:l)
>                               second (s,l) = (s,Second:l)
> subexps exp@(Tupl es t) 
>  = [(exp,[])] 
>     ++ concat [ map (tupelem n) $ subexps  $ es!!n | n <- [0..(length es-1)]]
>    where tupelem n (s,l) = (s, TupElem n:l)

> replaces :: Exp -> [Location] -> Exp -> Exp
> replaces e [] _ = e
> replaces e (loc:locs) r = replaces (replace e loc r) locs r


> replace :: Exp -> Location -> Exp -> Exp
> replace e [] r = r
> replace (Ap e1 e2 t) (First:ds) r = Ap (replace e1 ds r) e2 t
> replace (Ap e1 e2 t) (Second:ds) r = Ap e1 (replace e2 ds r) t
> replace (Tupl es t) (TupElem n:ds) r 
>    = Tupl (take n es ++ [replace (es!!n) ds r] ++ drop (n+1) es) t
> replace (Lam p e t) (Body:ds) r = Lam p (replace e ds r) t
> replace e l r = error ("invalid replace:\n" ++ "e=" ++ show e ++ "\n"
>                                             ++ "l=" ++ show l ++ "\n"
>                                             ++ "r=" ++ show r ++ "\n")

> replace' :: Exp -> Location -> Exp -> Exp
> replace' e [] r = r
> replace' (Ap e1 e2 t) (First:ds) r = Ap (replace e1 ds r) e2 t
> replace' (Ap e1 e2 t) (Second:ds) r = Ap e1 (replace e2 ds r) t
> replace' (Tupl es t) (TupElem n:ds) r 
>    = Tupl (take n es ++ [replace' (es!!n) ds r] ++ drop (n+1) es) t
> replace' (Lam p e t) (Body:ds) r = etaRed (Lam p (replace e ds r) t)
> replace' e l r = error ("invalid replace:\n" ++ "e=" ++ show e ++ "\n"
>                                             ++ "l=" ++ show l ++ "\n"
>                                             ++ "r=" ++ show r ++ "\n")

> twostepappsplit :: AppSplitter

> twostepappsplit fflex e t = 
>       if fflex then do x <- freshVar
>                        b <- liftSt $ abstracts x e t
>                        return (etaRed (lambda (x,exptype e) b),[])
>       else simpleappsplit fflex e t

 tracesplit str appsplitter fflex e t =
    trace (str ++ "(" ++ show fflex ++ ") " ++ showExp e ++ " -> " 
               ++ showExp t ++ "\n") $
       do (res,rules) <- appsplitter fflex e t
          trace ("res: " ++ showExp res ++ "\n") (return (res,rules))

> abstracts :: VarId -> Exp -> Exp -> [Exp]
> abstracts x e t = concat $ takeWhile (not.null) $
>                      [abstractssub n x e t | n <- [0..]]

> abstractssub :: Int -> VarId -> Exp -> Exp -> [Exp]

> abstractssub 0 x e t = [t]
> abstractssub (n+1) x e t = nub [c | b <- abstractssub n x e t, 
>                                     c <- abstract x e b]

> abstract :: VarId -> Exp -> Exp -> [Exp]

> abstract x e t = [replace' t loc r | (s,loc) <- subexps t,
>                                     not (x `appearsin` s),
>                                     r <- instanc x e s]
>            where appearsin x (LVar y _) = x==y
>                  appearsin x (PVar _ _) = False
>                  appearsin x (Con _ _) = False
>                  appearsin x (Lam _ b _) = appearsin x b
>                  appearsin x (Ap e1 e2 _) = appearsin x e1 || appearsin x e2
>                  appearsin x (Tupl es _) = or $ map (appearsin x) es

> instanc x e s = 
>    [foldl apl (LVar x (exptype e)) (apply phi xs) 
>                     | let xs = args e,
>                       phi <- simplematch (body e) s]

>    where simplematch p t = 
>                  map (funcom (thaw cs)) $
>                     evalSt ([],[]) $
>                       matches simpleappsplit [Rule (p',t')]
>               where bs = unboundlocals p
>                     cs = unboundlocals t
>                     p' = freeze bs p
>                     t' = freeze cs t
>          args (Lam (LamVar x) exp t) = PVar x (argtype t):args exp
>          args exp = []

>          body (Lam (LamVar x) exp _) = patternize x (body exp)
>          body exp = exp

>          patternize x (LVar y t) = if x==y then PVar y t else LVar y t
>          patternize x (PVar p t) = PVar p t
>          patternize x (Con c t) = Con c t
>          patternize x (Ap e1 e2 t) = Ap (patternize x e1) (patternize x e2) t
>          patternize x (Tupl es t) = Tupl (map (patternize x) es) t
>          patternize x (Lam p exp t) = Lam p (patternize x exp) t

Turn an appsplitter into one that can handle tuples.

> tuplifysplitter :: AppSplitter -> AppSplitter

We only modify the behaviour if the argument is an explicit tuple.

> tuplifysplitter appsplitter fflex (Tupl args _) term 
>     = do (xs,body,rules) <- foldM (onesplit fflex) ([],term,[]) args
>          return (etaRed $ tuplelambda (reverse xs) body,rules)

Parameters: is the function flexible
            (current list of parameters, current term, current rules)
            new argument to be split out

>   where onesplit fflex (xs,term',rules) arg = 

Get a new function part and list of rules for the current argument and term.

>                do (newfunc,newrules) 

Call tuplifysplitter appsplitter because of the possibility that the nested
argument might also be a tuple. Note that fflex is forced to True.

>                         <- tuplifysplitter appsplitter True arg term'

Get the parameter and body parts out from the function part obtained from
splitting.

>                   (x,newbody) <- lamsplit newfunc

Add the parameter and new rules to the appropriate lists.

>                   return (x:xs,newbody,newrules ++ rules)


>         lamsplit (Lam p b typ) = return ((p,argtype typ),b)
>         lamsplit e = do x <- freshVar
>                         let typ = argtype (exptype e)
>                         return ((LamVar x,typ),apl e (LVar x typ))

If it's not a tuple, just call the original splitter

> tuplifysplitter appsplitter fflex arg term = appsplitter fflex arg term
